<?php

/*
  +-------------------------------------------------------------------------------------------------+
  | ObsMania Plugin                                             
  +-------------------------------------------------------------------------------------------------+
  | Copyright (c) 2013 - 2014 ObsMania                        
  +-------------------------------------------------------------------------------------------------+
  | This plugin is written to send Obstacle records (and others informations) 
  | on a centralized server.
  | All records are displayed on the record widget whatever the server where 
  | it was recorded.
  | Records and Global Ranking are displayed on the Obsmania website
  | and soon directly from the plugin.
  | Map votes are saved on the cloud too (like MX), 
  | but you can disable this function if you prefer MX.
  | 
  | Informations displayed on the widgets are refreshed all 15 seconds (never if nobody on the server).
  |
  | A command can be used to import all Maniacontrol locals records on
  | the ObsMania Server (/omimport|omi) and can be only used by SA
  | 
  | The server administrator have to identify his server on the ObsMania Website
  | to connect the plugin to the remote Database
  |
  | IMPORTANT : The display function is a reuse of the recordWidget from Chris92 & TheM
  | 
  +-------------------------------------------------------------------------------------------------+
  | Author: MarlburroW <78195@supinfo.com>                                                               
  +-------------------------------------------------------------------------------------------------+
*/


namespace ObsMania;

use ManiaControl\Settings\SettingManager;
use ManiaControl\Callbacks\Callbacks;
use ManiaControl\Callbacks\CallbackListener;
use ManiaControl\Callbacks\CallbackManager;
use ManiaControl\Callbacks\TimerListner;
use ManiaControl\Commands\CommandListener;
use ManiaControl\ManiaControl;
use ManiaControl\Manialinks\ManialinkPageAnswerListener;
use ManiaControl\Players\Player;
use ManiaControl\Players\PlayerManager;
use ManiaControl\Maps\MapManager;
use ManiaControl\Plugins\Plugin;
use ManiaControl\Callbacks\Models\RecordCallback;
use ManiaControl\Manialinks\ManialinkManager;
use ManiaControl\Utils\Formatter;
use ManiaControl\Callbacks\TimerListener;
use ManiaControl\Admin\AuthenticationManager;
use FML\Controls\Control;
use FML\Controls\Frame;
use FML\Controls\Label;
use FML\Controls\Quad;
use FML\Controls\Quads\Quad_BgsPlayerCard;
use FML\Controls\Quads\Quad_Icons128x128_1;
use FML\Controls\Quads\Quad_BgRaceScore2;
use FML\ManiaLink;
use FML\Script\Features\Paging;


/**
 * 
 * @name ObsMania Plugin
 * @author MarlburroW
 *
 */
class ObsMania implements CallbackListener, CommandListener, ManialinkPageAnswerListener, TimerListener, Plugin {
    
    /*##########################################################################################*/
    /*################################## Plugin Constants ######################################*/
    /*##########################################################################################*/

    /* ManiaControl Metadata */

    const ID                                            = 66;
    const VERSION                                       = 0.21;
    const NAME                                          = 'ObsMania Plugin';
    const AUTHOR                                        = 'MarlBurroW';
    const DESCRIPTION                                   = 'ObsMania connects your server to a centralized database of obstacle records';

    /* ObsMania private settings */

    const OBSMANIA_URL                                  = 'http://obsmania.com';
    const CALLS_DEBUG                                   = false;   /* Enable call history in the ManiaControl shell output */
    const REFRESH_INTERVAL                              = 15; /* Define the time between each refresh (seconds)

    /* Text Settings (to long valuesto be added in the Maniacontrol settings system :( ) */

    const TEXT_MAP_FINISHED_TRAINING_MODE               = '$z$faoYour time : $o$n$FF0{{newtime}}';
    const TEXT_MAP_FINISHED_FIRST_TIME                  = '{{nickname}}$z$FAO finish the map for the first time and get the rank $o$n$FF0{{newrank}} $z$FAOwith$o$n $FF0{{newtime}}';
    const TEXT_MAP_FINISHED_NEW_RECORD                  = '{{nickname}}$z$FAO get the rank $o$n$FF0{{newrank}}$z$FAO with $o$n$FF0{{newtime}} $z$3F0(-{{timediff}})';
    const TEXT_MAP_FINISHED                             = '$z$FAO Your time : $o$n$FF0{{newtime}} $z$FAO| Your actual record : $o$n$FF0{{oldtime}} $z$f70| $o$n$F30(+{{timediff}})';
    const TEXT_VOTE_REFUSED                             = '$F30Please {{nickname}}$z$F30, finish the map before voting !';
    const TEXT_VOTE_1                                   = '{{nickname}}$z$39f hate this map and vote $F001';
    const TEXT_VOTE_2                                   = '{{nickname}}$z$39f doesn\'t really like this map and vote $F702';
    const TEXT_VOTE_3                                   = '{{nickname}}$z$39f find this map pretty good and vote $CF03.';
    const TEXT_VOTE_4                                   = '{{nickname}}$z$39f love this map and vote $8FO4';
    const TEXT_VOTE_5                                   = '{{nickname}}$z$39f find this map AMAZING and vote  $3F05';

    /* Widget Settings  */

    const MLID_RECORDS                                  = 'ml_obsmania_ranking';
    const SETTING_WIDGET_ENABLED                        = 'Enable/Disable Record Widget';
    const SETTING_WIDGET_LOCALRECORDING                 = 'Save records in local database';
    const SETTING_WIDGET_TITLE                          = 'Record Widget Title';
    const SETTING_WIDGET_POSX                           = 'Record Widget Position: X';
    const SETTING_WIDGET_POSY                           = 'Record Widget Position: Y';
    const SETTING_WIDGET_WIDTH                          = 'Record Widget Width';
    const SETTING_WIDGET_LINESCOUNT                     = 'Record Widget Displayed Lines Count';
    const SETTING_WIDGET_LINEHEIGHT                     = 'Record Widget Line Height';
    const SETTING_WIDGET_TOPCOUNT                       = 'Record Widget # of shown top records';       
    const SETTING_TRAINING_MODE                         = 'Record Widget Enable/disable training mode'; 

    const MLID_MAPINFO                                  = 'ml_obsmania_mapinfo';    
    const SETTING_MAPINFO_WIDGET_ENABLED                = 'Enable/Disable Map Info Widget';
    const SETTING_MAPINFO_WIDGET_POSX                   = 'Map Info Widget Position: X';
    const SETTING_MAPINFO_WIDGET_POSY                   = 'Map Info Widget Position: Y';
    const SETTING_MAPINFO_WIDGET_WIDTH                  = 'Map Info Widget Width';

    const MLID_SERVERSWITCH                             = 'ml_obsmania_serverswitch';   
    const SETTING_SERVERSWITCH_WIDGET_ENABLED           = 'Enable/Disable Server switch Widget';
    const SETTING_SERVERSWITCH_WIDGET_POSX              = 'Server switch Widget Position: X';
    const SETTING_SERVERSWITCH_WIDGET_POSY              = 'Server switch Widget Position: Y';
    const SETTING_SERVERSWITCH_WIDGET_WIDTH             = 'Server switch Widget Width';

    /* Actions */

    const ACTION_SHOW_RECORDSLIST                       = 'ObsMania.ShowRecordsList';
    const ACTION_SHOW_VOTES                             = 'ObsMania.ShowVotes';
    
    /* DataBase (for local recording when it's enabled) */

    const TABLE_RECORDS = 'mc_localrecords'; /* LocalRecordsPlugin record table for save local record */

    /*##################################################################################################*/
    /*################################## Plugin Properties #############################################*/
    /*##################################################################################################*/

    private $maniaControl = null;

    private $globalSettings = null;
    
    private $mapList = null;

    private $organizationInfo = null;

    private $currentMapInfo = null;

    private $currentMapRanking = null;

    private $callsEnabled = true; /* if false, all remote calls are blocked */

    private $serverlogin = '';

    /*##########################################################################################################*/
    /*################################## ManiaControl Plugin Implementation ####################################*/
    /*##########################################################################################################*/

    /**
     * Get plugin id
     *
     * @return int
     */
    public static function getId() {
        return self::ID;
    }

    /**
     * Get Plugin Name
     *
     * @return string
     */
    public static function getName() {
        return self::NAME;
    }

    /**
     * Get Plugin Version
     *
     * @return float
     */
    public static function getVersion() {
        return self::VERSION;
    }

    /**
     * Get Plugin Author
     *
     * @return string
     */
    public static function getAuthor() {
        return self::AUTHOR;
    }

    /**
     * Get Plugin Description
     *
     * @return string
     */
    public static function getDescription() {
        return self::DESCRIPTION;
    
    }

    public static function prepare(ManiaControl $maniaControl) {

        
    }


    /**
     * Load the plugin
     *
     * @param \ManiaControl\ManiaControl $maniaControl
     */
    public function load(ManiaControl $maniaControl) {

        date_default_timezone_set ('Europe/Paris');
        
        $this->maniaControl = $maniaControl;

        /* Plugin Settings */

        $maniaControl->settingManager->initSetting($this,self::makeServerPasswordSettingName($maniaControl),'');

        /* Widgets settings */

        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_ENABLED, true);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_LOCALRECORDING, false);      
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_TITLE, '$o$s$f70Obs$fffMania Records');
        $maniaControl->settingManager->initSetting($this, self::SETTING_TRAINING_MODE, false);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_POSX, 139.);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_POSY, 75);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_WIDTH, 40.);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_LINESCOUNT, 15);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_LINEHEIGHT, 4.);
        $maniaControl->settingManager->initSetting($this, self::SETTING_WIDGET_TOPCOUNT, 3);
    
        $maniaControl->settingManager->initSetting($this, self::SETTING_MAPINFO_WIDGET_ENABLED, true);
        $maniaControl->settingManager->initSetting($this, self::SETTING_MAPINFO_WIDGET_POSX, 139.);
        $maniaControl->settingManager->initSetting($this, self::SETTING_MAPINFO_WIDGET_POSY, 92);
        $maniaControl->settingManager->initSetting($this, self::SETTING_MAPINFO_WIDGET_WIDTH, 40);
        

        $maniaControl->settingManager->initSetting($this, self::SETTING_SERVERSWITCH_WIDGET_ENABLED, true);
        $maniaControl->settingManager->initSetting($this, self::SETTING_SERVERSWITCH_WIDGET_POSX, 139.);
        $maniaControl->settingManager->initSetting($this, self::SETTING_SERVERSWITCH_WIDGET_POSY, 8);
        $maniaControl->settingManager->initSetting($this, self::SETTING_SERVERSWITCH_WIDGET_WIDTH, 40);
        // Set properties

        $this->serverlogin = $this->maniaControl->server->getInfo()->login;

        /* Init */

    
        $this->refreshTime = time();

        if(!$this->isAuthConfigured()){
            $this->maniaControl->chat->sendMessageToAdmins(self::NAME . ' - First, you have to identify your server on the obsmania website and set the password in the plugin setting. Go to $l[http://obsmania.com]http://obsmania.com$l to create your server.');
            return;
        }
      

        $this->initCalls();

        
        /* Commands */

        $maniaControl->commandManager->registerCommandListener(array('omi', 'omimport'), $this, 'importManiaControlLocalData',false, 'Upload ManiaControl\'s local data to ObsMania server');

        //Callbacks

        $this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERCONNECT, $this, 'handlePlayerConnect');
        $this->maniaControl->callbackManager->registerCallbackListener(PlayerManager::CB_PLAYERDISCONNECT, $this, 'handlePlayerDisconnect');
        $this->maniaControl->callbackManager->registerCallbackListener(CallbackManager::CB_MP_MAPLISTMODIFIED, $this, 'handleMapListModified');
        $this->maniaControl->callbackManager->registerCallbackListener(Callbacks::BEGINMAP, $this, 'handleBeginMap');
        $this->maniaControl->callbackManager->registerCallbackListener(RecordCallback::FINISH, $this, 'handleFinishCallback');
        $this->maniaControl->callbackManager->registerCallbackListener(SettingManager::CB_SETTING_CHANGED, $this, 'handleSettingChanged');
        $this->maniaControl->callbackManager->registerCallbackListener(CallbackManager::CB_MP_PLAYERMANIALINKPAGEANSWER, $this, 'handleManialinkPageAnswer');
        $this->maniaControl->callbackManager->registerCallbackListener(CallbackManager::CB_MP_PLAYERCHAT, $this, 'handlePlayerChat');
        
        $this->maniaControl->timerManager->registerTimerListening($this, 'handle1Second', self::REFRESH_INTERVAL * 1000);

       

    }

    /**
     * Unload the plugin and its resources
     */

    public function unload() {

        $this->hideAll();

        $this->globalSettings = null;
    
        $this->mapList = null;

        $this->currentMapInfo = null;

        $this->currentMapRanking = null;

        $this->organizationInfo = null;
    }




    /*###########################################################################################################*/
    /*######################################### Remote Access Functions #########################################*/
    /*###########################################################################################################*/


    /* INFO : All remote calls to Obsmania are async and use closure functions in a callback */

    /**
     * All required calls to work
     * @return void
     */
    public function initCalls(){

        $this->maniaControl->chat->sendMessageToAdmins('ObsMania Plugin - Refreshing all data...');

        $this->updateServerInfo();
        $this->getOrganizationInfo();
        $this->getGlobalSettings();
        $this->createOrUpdatePlayers();
        $this->updateMapList();
        
        $this->getCurrentMapInfo();
        $this->getCurrentMapRanking();

    }

    /**
     * Send some informations about the current server to the ObsMania server
     * @return void
     */
    public function updateServerInfo(){

        $serverInfos = $this->maniaControl->server->getInfo();


        $serverName = $serverInfos->nickName;

        $data = [
            'serverName' => $serverName,
            'maxPlayers' =>  $this->maniaControl->client->getMaxPlayers()['NextValue'],
            'playerCount' => $this->maniaControl->playerManager->getPlayerCount(true),
            'maxSpectators' => $this->maniaControl->client->getMaxSpectators()['NextValue'],
            'spectatorCount' => $spectatorCount = $this->maniaControl->playerManager->getSpectatorCount(),
        ];

        $this->remoteCall('updateServerInfo',$data, function ($data){

        });

    }

    /**
     * Get some informations about the organization of the server (and other servers info for the server switch widget)
     * @return void
     */
    public function getOrganizationInfo(){

        $this->remoteCall('getOrganizationInfo',function($organization){

            $this->organizationInfo = $organization;

            $this->displayServerSwitchWidget();

           
        });


    }

    /**
     * Get controlled settings from the Obsmania server (not used yet)
     * @return void
     */
    public function getGlobalSettings(){

        $this->remoteCall('getGlobalSettings', function ($data){});

    }

    /**
     * Send informations (nickname,login) about connected players and create them in the ObsMania database
     * @return void
     */
    public function createOrUpdatePlayers(){

        $players = $this->getPlayerList();

        $this->remoteCall('createOrUpdatePlayers',$players,function($data){

            $this->updatePlayerList();
        });
    }

    /**
     * Send informations (nickname,login) about a specific connected player and create them in the ObsMania database
     * @return void
     */
    public function createOrUpdatePlayer($playerlogin){

        $player = $this->getPlayer($playerlogin);

        $this->remoteCall('createOrUpdatePlayer',$player, function ($data){});
    }

    /**
     * Send a list of the connected players to the ObsMania server
     * @return void
     */
    public function updatePlayerList(){

        $players = $this->getPlayerList();

        $this->remoteCall('updatePlayerList',$players, function ($data){});
    }

    /**
     * Send the map playlist to the ObsMania server
     * @return void
     */
    public function updateMapList(){
        $maps = $this->getMapList();

        $this->remoteCall('updateMapList',$maps,function($data){

           $this->getServerMapList();
           $this->updateCurrentMap();

        });
    }

    /**
     * Get all maps in the server playlist from ObsMania with votes informations
     * @return void
     */
    public function getServerMapList(){

        $this->remoteCall('getServerMapList',function($data){
            $this->mapList = $data;

        });

    }

    /**
     * Send the current map UID to the ObsMania server
     * @return void
     */
    public function updateCurrentMap(){

        $map = $this->getCurrentMap();

        $data = array('challengeid'=>$map->uid);
        $this->remoteCall('updateCurrentMap',$data,function($data){

            

        });
    }

    /**
     * Get informations about the current map from the Obsmaina Server and display it on the MapInfo widget
     * @return void
     */
    public function getCurrentMapInfo(){

        $data = array(
            'challengeid' => $map = $this->getCurrentMap()->uid
        );
        $this->remoteCall('getMapInfo',$data,function($data){

            $this->currentMapInfo = $data;
            $this->displayMapInfdoWidget();
        });
        
    }

    /**
     * Get the current Map ranking from the ObsMania server and display it on the Record Widget
     * @return void
     */
    public function getCurrentMapRanking(){

        $data = array(
            'challengeid' => $map = $this->getCurrentMap()->uid
        );

        $this->remoteCall('getMapRanking',$data,function($data){

            $this->currentMapRanking = $data->ranking;

            $this->displayRecordWidget();
            
        });


    }

    /**
     * Get the current Map ranking from the ObsMania server and display it on the Record Widget
     * @return void
     */
    public function voteMap($player,$vote){

        $vote = array(
                'playerlogin' => $player->login,
                'vote' => $vote,
                'challengeid' => $this->getCurrentMap()->uid,
        );

        $this->remoteCall('voteMap',$vote,function($data) use ($vote,$player) {

            if($data->result == 'error' && $data->code == 'map-not-finished')$this->maniaControl->chat->sendError($this->renderText(ObsMania::TEXT_VOTE_REFUSED,array('nickname' => $player->nickname)));
            else{
                
              
                switch($vote['vote']){
                    case 1:
                        $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_VOTE_1,array('nickname' => $player->nickname)));
                    break;
                    case 2:
                        $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_VOTE_2,array('nickname' => $player->nickname)));
                    break;
                    case 3:
                        $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_VOTE_3,array('nickname' => $player->nickname)));
                    break;
                    case 4:
                        $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_VOTE_4,array('nickname' => $player->nickname)));
                    break;
                    case 5:
                        $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_VOTE_5,array('nickname' => $player->nickname)));
                    break;
                }
                $this->getCurrentMapInfo();
                $this->getServerMapList();   
            }
            
        });

    }

    public function writeRecord($playerlogin,$time,$challengeid){

        $data = array(
            'challengeid' => $challengeid,
            'playerlogin' => $playerlogin,
            'time' =>$time,
        );

        

        $this->remoteCall('createOrUpdateRecord',$data,function($record){
            
             $player = $this->getPlayer($record->playerlogin);
            if($this->maniaControl->settingManager->getSettingValue($this, self::SETTING_TRAINING_MODE)){

                $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_MAP_FINISHED_TRAINING_MODE, [
                    'nickname' => $player['NickName'], 
                    'newtime'  =>  $this->toSMTime($record->time),
                ]),$record->playerlogin);

            }elseif($record->isFirstRecord){

                $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_MAP_FINISHED_FIRST_TIME, [
                    'nickname' => $player['NickName'], 
                    'newrank'  => $record->newRank, 
                    'newtime'  =>  $this->toSMTime($record->time)
                ]));

            }elseif($record->isNewRecord){

                $gain = $this->toSMTime($record->lastTime - $record->time);

                $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_MAP_FINISHED_NEW_RECORD, [
                    'nickname' => $player['NickName'], 
                    'newrank'  => $record->newRank, 
                    'oldrank'  => $record->oldRank, 
                    'newtime'  =>  $this->toSMTime($record->time),
                    'timediff' => $gain,
                    'oldtime' => $this->toSMTime($record->lastTime),
                ]));
                
            
            }else{

                $gain = $this->toSMTime($record->time - $record->lastTime);
                $this->maniaControl->chat->sendInformation($this->renderText(ObsMania::TEXT_MAP_FINISHED, [
                    'nickname' => $player['NickName'], 
                    'oldtime' =>  $this->toSMTime($record->lastTime), 
                    'newtime' =>  $this->toSMTime($record->time),
                    'timediff' => $gain
                ]),$record->playerlogin);
                
            }

        });
        
        $this->getCurrentMapRanking();
    
    }

    public function refreshInfo(){

        if($this->maniaControl->playerManager->getPlayerCount(true) > 0 ){ /* Refresh infos only if players connected */

            $this->getCurrentMapRanking();
            $this->getCurrentMapInfo();
            $this->getOrganizationInfo();

        }
    
    }


    /*###########################################################################################################*/
    /*######################################### Events Callbacks ################################################*/
    /*###########################################################################################################*/

    public function handlePlayerConnect(Player $player) {

        $player = $this->getPlayer($player->login);

        $this->updateServerInfo();

        $this->remoteCall('createOrUpdatePlayer',$player,function(){
               $this->updatePlayerList();
        });
        $this->displayRecordWidget();

    }

    public function handlePlayerDisconnect(Player $player) {
       $this->updateServerInfo();
       $this->updatePlayerList();

    }

    public function handleMapListModified(){

       $this->updateMapList();

    }

    public function handleBeginMap(){
      
        $this->getCurrentMapInfo();
        $this->getCurrentMapRanking();
        $this->updateCurrentMap();
    }

    public function handleFinishCallback(RecordCallback $callback){

        $this->writeRecord($callback->player->login, $callback->time, $this->getCurrentMap()->uid);

        /* Save to local database if setting enabled */
        
        $saveLocalRecord = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_LOCALRECORDING);
        if($saveLocalRecord)$this->writeLocalRecord($callback->player->index,$callback->time, $this->getCurrentMap()->index);
    }
    public function handleSettingChanged(\ManiaControl\Settings\Setting $setting) {
        switch($setting->setting){
            case self::makeServerPasswordSettingName($this->maniaControl):
                if(!$this->isAuthConfigured()){
                    $this->maniaControl->chat->sendMessageToAdmins(self::NAME . ' - First, you have to identify your server on the obsmania website and set the password in the plugin setting. Go to $l[http://obsmania.com]http://obsmania.com$l to create your server.');
       
                }
                $this->hideAll();
                $this->initCalls();

            break;
            case self::SETTING_TRAINING_MODE:
                if($setting->value == 1){
                     $this->maniaControl->chat->sendInformation(self::NAME . ' - Training mode ON');
                }else{
                     $this->maniaControl->chat->sendInformation(self::NAME . ' - Training mode OFF');
                }
            break;
       }
        $this->callsEnabled = true;
        $this->displayAllWidgets();
        
       

    }

    public function handleManialinkPageAnswer(array $callback) {

        $actionId = $callback[1][2];
         
        $login  = $callback[1][1];
        $player = $this->maniaControl->playerManager->getPlayer($login);

        if ($actionId === self::ACTION_SHOW_RECORDSLIST) {
            $this->showRecordsList(array(), $player);
        }
        if ($actionId === self::ACTION_SHOW_VOTES) {
            $this->ShowVotes(array(), $player);
        }

       
    }


    public function handle1Second(){
        $this->refreshInfo();
    }



    /**
     * Vote callback, send player a vote to Obsmania using the voteMap method
     * @return void
     */
    public function handlePlayerChat(array $chatCallback) {

        $login  = $chatCallback[1][1];
        $player = $this->maniaControl->playerManager->getPlayer($login);

        if (!$player) {
            return;
        }

        $message = $chatCallback[1][2];
        
        switch($message){
            case '--':
                $vote = 1;
                $this->voteMap($player, $vote);
            break;
            case '-':
                $vote = 2;
                $this->voteMap($player, $vote);
            break;
            case '-+':
                $vote = 3;
                $this->voteMap($player, $vote);
            break;
            case '+-':
                $vote = 3;
                $this->voteMap($player, $vote);
            break;
            case '+':
                $vote = 4;
                $this->voteMap($player, $vote);
            break;
            case '++':
                $vote = 5;
                $this->voteMap($player, $vote);
            break;
        }

        
    }

    /*###########################################################################################################*/
    /*######################################### Display Functions ###############################################*/
    /*###########################################################################################################*/

    /**
     * Display all widgets for all players
     * @return void
     */
    public function displayAllWidgets(){
        $this->displayMapInfdoWidget();
        $this->displayServerSwitchWidget();
        $this->displayRecordWidget();
    }

    /**
     * Display the Server Switch Widget for all players
     * @return void
     */
    public function displayServerSwitchWidget() {

        if ($this->maniaControl->settingManager->getSettingValue($this, self::SETTING_SERVERSWITCH_WIDGET_ENABLED)) {


            foreach($this->maniaControl->playerManager->getPlayers() as $player) {

                $this->buildServerSwitchWidget($player);

            }

        }else{

            $this->maniaControl->manialinkManager->hideManialink(self::MLID_SERVERSWITCH);

        }
        
        
    }

    /**
     * Display the Records Widget for all players
     * @return void
     */
    public function displayRecordWidget() {
        if ($this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_ENABLED)) {

            foreach($this->maniaControl->playerManager->getPlayers() as $player) {

                $this->buildRecordWidget($player);

            }

        }else{

            $this->maniaControl->manialinkManager->hideManialink(self::MLID_RECORDS);

        }
        
        
    }


    /**
     * Display the MapInfo Widget for all players
     * @return void
     */
    public function displayMapInfdoWidget(){

        if ($this->maniaControl->settingManager->getSettingValue($this, self::SETTING_MAPINFO_WIDGET_ENABLED)) {

            foreach($this->maniaControl->playerManager->getPlayers() as $player) {

                $this->buildMapInfoWidget($player);

            }
        }else{

            $this->maniaControl->manialinkManager->hideManialink(self::MLID_MAPINFO);

        }

    }

    /**
     * Build the server switch widget and display it to a player
     * @param Player $player 
     * @return void
     */

    private function buildServerSwitchWidget(Player $player){

        $pos_x        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_SERVERSWITCH_WIDGET_POSX);
        $pos_y        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_SERVERSWITCH_WIDGET_POSY);
        $width        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_SERVERSWITCH_WIDGET_WIDTH);
        $mlid         = self::MLID_SERVERSWITCH;
        

        $manialink = new ManiaLink($mlid);
        $frame     = new Frame();
        $manialink->add($frame);
        $frame->setPosition($pos_x, $pos_y);

        $backgroundQuad = new Quad();
        $frame->add($backgroundQuad);
        $backgroundQuad->setVAlign(Control::TOP);
        $adjustOuterBorder = false;

        $height            = 10. + (3 * count($this->organizationInfo->servers));
        $backgroundQuad->setSize($width * 1.2, $height);
        $backgroundQuad->setStyles('Bgs1InRace', 'BgList');
        
        $titleBackgroundQuad = new Quad();
        $frame->add($titleBackgroundQuad);
        $titleBackgroundQuad->setVAlign(Control::TOP);
        
        $titleBackgroundQuad->setY(-0.8);
        $titleBackgroundQuad->setX(0);
        $titleBackgroundQuad->setZ(2);
    


        $titleLabel = new Label();
        $frame->add($titleLabel);
        $titleLabel->setHAlign(Control::CENTER);
        $titleLabel->setPosition(0, -3);
        $titleLabel->setWidth($width);
        $titleLabel->setTextSize(1);
        $labelStyle   = $this->maniaControl->manialinkManager->styleManager->getDefaultLabelStyle();
        $titleLabel->setStyle($labelStyle);
        $titleLabel->setText($this->organizationInfo->name . '$z Servers');
        $titleLabel->setZ(3);
        $titleLabel->setTranslate(true);


        $topBackgroundQuad = new Quad();
        $frame->add($topBackgroundQuad);
        $topBackgroundQuad->setVAlign(Control::TOP);
        $adjustOuterBorder = false;
        $height            = (3 * count($this->organizationInfo->servers))+2;
        $topBackgroundQuad->setSize($width * 1.2, $height);
        $topBackgroundQuad->setY(-5.5);
        $topBackgroundQuad->setStyles('Bgs1InRace', 'BgCardList');

        $y = -8;


        foreach($this->organizationInfo->servers  as $server){

            $serverFrame = new Frame();
            $serverFrame->setPosition(0, $y);

            $serverNameLabel = new Label();
            $serverFrame->add($serverNameLabel);
            $serverNameLabel->setHAlign(Control::LEFT);
            $serverNameLabel->setX(-20);
        
            $serverNameLabel->setTextSize(0.9);
            $serverNameLabel->setText($server->name);
           
            $serverNameLabel->setUrl('maniaplanet://#join=' . $server->serverlogin) ;
            $serverNameLabel->setTextEmboss(true);

            $playerCountLabel = new Label();
            $serverFrame->add($playerCountLabel);
            $playerCountLabel->setHAlign(Control::RIGHT);
            $playerCountLabel->setX($width * 0.49);
            
            $playerCountLabel->setTextSize(0.9);
            $text = $server->playercount . ' / ' . $server->maxplayers;
            $playerCountLabel->setText($text);

        
            $frame->add($serverFrame);
            $y -= 3;

        }


        $this->maniaControl->manialinkManager->sendManialink($manialink, $player->login);


    }

    /**
     * Build the map info widget and display it for the specified player
     * @param  Player $player 
     * @return void
     */
    private function buildMapInfoWidget(Player $player) {
        
        $pos_x        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_MAPINFO_WIDGET_POSX);
        $pos_y        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_MAPINFO_WIDGET_POSY);
        $width        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_MAPINFO_WIDGET_WIDTH);
        $mlid         = self::MLID_MAPINFO;
        
        $labelStyle   = $this->maniaControl->manialinkManager->styleManager->getDefaultLabelStyle();


        $leftSide = ($pos_x < 0) ? true : false;

        $manialink = new ManiaLink($mlid);
        $frame     = new Frame();
        $manialink->add($frame);
        $frame->setPosition($pos_x, $pos_y);

        $backgroundQuad = new Quad();
        $frame->add($backgroundQuad);
        $backgroundQuad->setVAlign(Control::TOP);
        
        $height            = 17;
        $backgroundQuad->setSize($width * 1.2, $height);
        $backgroundQuad->setStyles('Bgs1InRace', 'BgList');
        $backgroundQuad->setAction(self::ACTION_SHOW_VOTES);
    
        $titleLabel = new Label();
        $frame->add($titleLabel);
        $titleLabel->setHAlign(Control::CENTER);
        $titleLabel->setPosition(0, -5);
        $titleLabel->setWidth($width);
        // $titleLabel->setStyle($labelStyle);
        $titleLabel->setTextSize(1);
        $titleLabel->setText($this->currentMapInfo->challengename);
        $titleLabel->setZ(3);
        $titleLabel->setTranslate(true);


        $mapper = $this->currentMapInfo->mapper ? $this->currentMapInfo->mapper->nickname . ' $z$FFF(' . $this->currentMapInfo->mapper_playerlogin . ')' : $this->currentMapInfo->mapper_playerlogin;
        $mapperLabel = new Label();
        $frame->add($mapperLabel);
        $mapperLabel->setHAlign(Control::CENTER);
        $mapperLabel->setPosition(0, -7.5);
        $mapperLabel->setWidth($width); 
        $mapperLabel->setTextSize(0.9);
        $mapperLabel->setText('By ' . $mapper );
        $mapperLabel->setZ(3);
        $mapperLabel->setTranslate(true);


        
        
        $titleBackground = new Quad();
        $frame->add($titleBackground);
        $titleBackground->setHAlign(Control::CENTER);
        $titleBackground->setPosition(0, -13);
        $height            = 8;
        $titleBackground->setSize($width * 1.2, $height);
        $titleBackground->setStyles('Bgs1InRace', 'BgList');

        $voteColor = $this->getVoteColor($this->currentMapInfo->rating);
        $playersVoteCount = 0;
        foreach($this->currentMapInfo->playersvotes as $vote){
            $playersVoteCount++;
        }

        $ratingText = $playersVoteCount > 0 ? 'Map rating : ' . $voteColor . $this->formatDecimal($this->currentMapInfo->rating) : 'Map rating : -,--';

        $ratingLabel = new Label();
        $frame->add($ratingLabel);
        $ratingLabel->setHAlign(Control::CENTER);
        $ratingLabel->setPosition(0, -12);
        $ratingLabel->setWidth($width); 
        $ratingLabel->setTextSize(1);
        
        $ratingLabel->setText($ratingText );
        $ratingLabel->setZ(3);
        $ratingLabel->setTranslate(true);

        if($playersVoteCount == 0) $voteCountText = 'No votes yet';
        elseif($playersVoteCount == 1)  $voteCountText =  '1 vote';
        else $voteCountText = $playersVoteCount . ' votes';

        
        $voteCount = new Label();
        $frame->add($voteCount);
        $voteCount->setHAlign(Control::CENTER);
        $voteCount->setPosition(0, -15);
        $voteCount->setWidth($width);   
        $voteCount->setTextSize(0.9);
        $voteCount->setText($voteCountText);

        $this->maniaControl->manialinkManager->sendManialink($manialink, $player->login);

    }

    /**
     * Build the records table and display it to the specified player
     * @param Player $player 
     * @return void
     */
    private function buildRecordWidget(Player $player) {
        

        $currentPlayers = $this->maniaControl->playerManager->getPlayers();
        $currentPlayerLogins = array();

        foreach($currentPlayers as $currentPlayer) {
            $currentPlayerLogins[] = $currentPlayer->login;
        }


        $labelStyle   = $this->maniaControl->manialinkManager->styleManager->getDefaultLabelStyle();
        $lineHeight   = 4.;


        $title        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_TITLE);
        $pos_x        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_POSX);
        $pos_y        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_POSY);
        $width        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_WIDTH);
        $lines        = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_LINESCOUNT);
        $topcount     = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_WIDGET_TOPCOUNT);
        $training_mode     = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_TRAINING_MODE);

        $mlid         = self::MLID_RECORDS;
        

        $leftSide = ($pos_x < 0) ? true : false;

        $manialink = new ManiaLink($mlid);
        $frame     = new Frame();
        $manialink->add($frame);
        $frame->setPosition($pos_x, $pos_y);

        $backgroundQuad = new Quad();
        $frame->add($backgroundQuad);
        $backgroundQuad->setVAlign(Control::TOP);
        $adjustOuterBorder = false;
        $height            = 7. + ($lines * $lineHeight);
        $backgroundQuad->setSize($width * 1.2, $height);
        $backgroundQuad->setStyles('Bgs1InRace', 'BgList');
        $backgroundQuad->setAction(self::ACTION_SHOW_RECORDSLIST);

        $titleBackgroundQuad = new Quad();
        $frame->add($titleBackgroundQuad);
        $titleBackgroundQuad->setVAlign(Control::TOP);
        $titleBackgroundQuad->setSize($width * 1.17, $lineHeight);
        $titleBackgroundQuad->setY(-0.8);
        $titleBackgroundQuad->setX(0);
        $titleBackgroundQuad->setZ(2);
    


        $titleLabel = new Label();
        $frame->add($titleLabel);
        $titleLabel->setHAlign(Control::CENTER);
        $titleLabel->setPosition(0, $lineHeight * -0.72);
        $titleLabel->setWidth($width);
        $titleLabel->setStyle($labelStyle);
        $titleLabel->setTextSize(1);
        $titleLabel->setText($title);
        $titleLabel->setZ(3);
        $titleLabel->setTranslate(true);


        $topBackgroundQuad = new Quad();
        $frame->add($topBackgroundQuad);
        $topBackgroundQuad->setVAlign(Control::TOP);
        $adjustOuterBorder = false;
        $height            = ($topcount * $lineHeight)+1;
        $topBackgroundQuad->setSize($width * 1.2, $height);
        $topBackgroundQuad->setY(-5.5);
        $topBackgroundQuad->setStyles('Bgs1InRace', 'BgCardList');

        
        $records = $this->currentMapRanking;
    

        $newRecords = array();
        foreach($records as $record) {
            
            $newRecords[] = $record;
        }
        $records = $newRecords; 

        if ($records == null || !is_array($records) || $records == false) {

            $y = -8. - 0 * $lineHeight;
            $recordFrame = $this->fillRecordLine(null, $y, $width, $lineHeight, $player, $currentPlayerLogins, $leftSide);
            $frame->add($recordFrame);

            $this->maniaControl->manialinkManager->sendManialink($manialink, $player->login);
            return;
        }
        $playerHasRecord = null;


        

        foreach($records as $record) {
            if($record->playerlogin == $player->login) {
                $playerHasRecord = $record;
            }
            
        }

            


        $index = 0;
        for($i = 0; $i < $topcount; $i++) {
            $y = -8. - $index * $lineHeight;

            if(isset($records[$i])) {
                if($training_mode && $records[$i]->playerlogin != $player->login)continue;
                $recordFrame = $this->fillRecordLine($records[$i], $y, $width, $lineHeight, $player, $currentPlayerLogins, $leftSide, $training_mode);
                $frame->add($recordFrame);
                $index++;
            }
        }



        $numberOfRecords = count($records);
        if($playerHasRecord == null) { // Player has no local record
            $needRecords = ($lines-$topcount)-1; // Add an additional -. --- -:--.--- line, so one less line needed
            $end = $numberOfRecords;
            $start = (($end - $needRecords) < ($topcount+1)) ? ($topcount) : ($end - $needRecords);
            $addEmptyLine = true;
        } else { // Player has a local record
            $needRecords = ($lines-$topcount);

            if($playerHasRecord->rank <= ($topcount+1)) { // Player has local record in TOP X
                $start = $topcount;
                $end = ($numberOfRecords < $needRecords) ? $numberOfRecords : ($start + $needRecords);
            } else { // Player doesn't have a record in the TOP X
                $recordsBelow = $numberOfRecords-$playerHasRecord->rank;
                $recordsAbove = $numberOfRecords-$topcount-1;
                $halfNeededBelow = floor(($needRecords-1)/2);
                $halfNeededAbove = ceil(($needRecords-1)/2);

                if($playerHasRecord->rank-$halfNeededAbove < ($topcount+1)) { // If requesting records into TOP X (or -1)
                    $start = $topcount; // Set start to first record after TOP X
                } elseif($playerHasRecord->rank-$halfNeededAbove > $topcount) { // If enough records above
                    $possibleExtraAbove = ($playerHasRecord->rank-$topcount-$halfNeededAbove < 0) ? 0 : $playerHasRecord->rank-$topcount-$halfNeededAbove;
                    $start = $playerHasRecord->rank-$halfNeededAbove-1;

                    if(!isset($records[$playerHasRecord->rank+$halfNeededBelow-1])) { // Not enough records below
                        //$tooLess = count($records)-($playerHasRecord->rank+$halfNeededBelow);
                        $haveBelow = count($records)-$playerHasRecord->rank;
                        $tooLess = ($lines-$topcount)-$halfNeededAbove-$haveBelow;
                        if($tooLess <= $possibleExtraAbove) {
                            $start = $playerHasRecord->rank-$halfNeededAbove-$tooLess;
                        } else {
                            $start = $playerHasRecord->rank-$halfNeededAbove-$possibleExtraAbove;
                        }
                    }
                }

                $end = ($numberOfRecords < $needRecords) ? $numberOfRecords : ($start + $needRecords);
            }
        }

        if(isset($start) && isset($end)) {
            for($i = $start; $i < $end; $i++) {
                $y = -8. - $index * $lineHeight;

                if(isset($records[$i])) {

                    
                    if($training_mode && $records[$i]->playerlogin != $player->login) continue;

                    $recordFrame = $this->fillRecordLine($records[$i], $y, $width, $lineHeight, $player, $currentPlayerLogins, $leftSide, $training_mode);
                    $frame->add($recordFrame);
                    $index++;
                } else {
                    break;
                }
            }
        }

        if(isset($addEmptyLine) && $addEmptyLine) {
            $y = -8. - $index * $lineHeight;
            $recordFrame = $this->fillRecordLine(null, $y, $width, $lineHeight, $player, $currentPlayerLogins, $leftSide, $training_mode);
            $frame->add($recordFrame);

            $index++;
        }

        $this->maniaControl->manialinkManager->sendManialink($manialink, $player->login);
    }


    /**
     * Process a line and add it to the records table
     * @param  $record 
     * @param  $y 
     * @param  $width 
     * @param  $lineHeight 
     * @param  $currentPlayer 
     * @param  $currentPlayerLogins 
     * @param  $leftSide 
     * @param  $training_mode 
     * @return 
     */

    private function fillRecordLine($record, $y, $width, $lineHeight, $currentPlayer, $currentPlayerLogins, $leftSide, $training_mode = false) {
        

        $recordFrame = new Frame();
        $recordFrame->setPosition(0, $y);

        $rankLabel = new Label();
        $recordFrame->add($rankLabel);
        $rankLabel->setHAlign(Control::RIGHT);
        $rankLabel->setX($width * -0.39);
        $rankLabel->setSize($width * 0.1, $lineHeight);
        $rankLabel->setTextSize(1);
    
        $linecolor = '';

          


        if($record == null) {
            $rankLabel->setText('');
        } else {

            if($training_mode){
                $linecolor = '$39F';
            }elseif($record->rank == 1){
                $linecolor = '$CC0';
            }elseif($record->rank == 2){
                $linecolor = '$999';
            }elseif($record->rank == 3){
                $linecolor = '$963';
            }elseif($record->playerlogin == $currentPlayer->login){
                $linecolor = '$39F';
            }elseif(in_array($record->playerlogin, $currentPlayerLogins)){
                 $linecolor = '$9CF';
            }elseif($record->serverlogin != $this->serverlogin){
                 $linecolor = '$CCC';
            }else{
                $linecolor = '$FFF';
            }

            if( $training_mode ) $rank = '?';
            else $rank =  $record->rank;

            $rankLabel->setText($linecolor . $rank.'.');


           
            
        }

        $rankLabel->setStyle('TextCardSmallScores2');

        $nameLabel = new Label();
        $recordFrame->add($nameLabel);
        $nameLabel->setHAlign(Control::LEFT);
        $nameLabel->setX($width * -0.38);
        $nameLabel->setSize($width * 0.6, $lineHeight);
        $nameLabel->setTextSize(1);
        if($record == null) {
            $nameLabel->setText($currentPlayer->nickname);
        } else {
            if(!isset($record->nickname)) {
                $nameLabel->setText($record->playerlogin);
            }
            else{
                $nameLabel->setText($record->nickname);
            }
        }
        $nameLabel->setTextEmboss(true);

        $timeLabel = new Label();
        $recordFrame->add($timeLabel);
        $timeLabel->setHAlign(Control::RIGHT);
        $timeLabel->setX($width * 0.49);
        $timeLabel->setSize($width * 0.25, $lineHeight);
        $timeLabel->setTextSize(1);

        

        if($record == null) {
            $timeLabel->setX($width * 0.45);
            $text = $linecolor . '-:--:---';
        } else {
            $text = $linecolor;
        
            $text .= $this->toSMTime($record->time);
         
        }
        $timeLabel->setText($text);

        if($record != null) {
            
                $timeLabel->setStyle('TextTitle2');
            
        }

        if($record != null) {
            $positionX = $width * -0.53;
            if($leftSide) $positionX = $width * 0.535;

            if($record->playerlogin == $currentPlayer->login) {
                $showOwnIcon = new Quad_Icons128x128_1();
                $recordFrame->add($showOwnIcon);
                $showOwnIcon->setSize(3, 3);
                $showOwnIcon->setY(-0.1);
                $showOwnIcon->setX($positionX);
                $showOwnIcon->setZ(2);
                $showOwnIcon->setSubStyle($showOwnIcon::SUBSTYLE_ChallengeAuthor);
            } elseif(in_array($record->playerlogin, $currentPlayerLogins)) {
                $showOthersIcon = new Quad_Icons128x128_1();
                $recordFrame->add($showOthersIcon);
                $showOthersIcon->setSize(3, 3);
                $showOthersIcon->setY(-0.05);
                $showOthersIcon->setX($positionX);
                $showOthersIcon->setZ(2);
                $showOthersIcon->setSubStyle($showOthersIcon::SUBSTYLE_Solo);
            }
        }

        return $recordFrame;
    }

    /**
     * Display the vote table
     * @param type array $chat 
     * @param type Player $player 
     * @return void
     */
    public function  ShowVotes(array $chat, Player $player) {
        $width  = $this->maniaControl->manialinkManager->styleManager->getListWidgetsWidth();
        $height = $this->maniaControl->manialinkManager->styleManager->getListWidgetsHeight();

        $votes = $this->currentMapInfo->playersvotes;

        // create manialink
        $maniaLink = new ManiaLink(ManialinkManager::MAIN_MLID);
        $script    = $maniaLink->getScript();
        $paging    = new Paging();
        $script->addFeature($paging);

        // Main frame
        $frame = $this->maniaControl->manialinkManager->styleManager->getDefaultListFrame($script, $paging);

        $maniaLink->add($frame);

        // Start offsets
        $posX = -$width / 2;
        $posY = $height / 2;

        // Predefine Description Label
        $descriptionLabel = $this->maniaControl->manialinkManager->styleManager->getDefaultDescriptionLabel();

        $frame->add($descriptionLabel);

        // Headline

        $headFrame = new Frame();
        $frame->add($headFrame);
        $headFrame->setY($posY - 5);

        $array = array(
            'NickName' => $posX + 5, 
            'Login' => $posX + 30,
            'Vote' => $posX + 100, 
            'Date' => $posX + 120, 
        );

        $this->maniaControl->manialinkManager->labelLine($headFrame, $array);
        $index     = 0;
        $posY      = $height / 2 - 10;
        $pageFrame = null;
        foreach ($votes as $vote) {


            if ($index % 15 === 0) {
                $pageFrame = new Frame();
                $frame->add($pageFrame);
                $posY = $height / 2 - 10;
                $paging->addPage($pageFrame);
            }

            $voteFrame = new Frame();
            $pageFrame->add($voteFrame);

            if ($index % 2 !== 0) {
                $lineQuad = new Quad_BgsPlayerCard();
                $voteFrame->add($lineQuad);
                $lineQuad->setSize($width, 4);
                $lineQuad->setSubStyle($lineQuad::SUBSTYLE_BgPlayerCardBig);
                $lineQuad->setZ(0.001);
            }

            if (strlen($vote->nickname) < 2) {
                $vote->nickname = $vote->login;
            }
            $currentPlayer = false;
            if($vote->playerlogin == $player->login) $currentPlayer = true;

            
            $array = array(
                $vote->nickname => $posX + 5,
                ($currentPlayer ? '$FF0' : '$FFF') . $vote->playerlogin => $posX + 30,
                $this->getVoteColor($vote->value) . $this->formatDecimal($vote->value) => $posX + 100,
                 $this->timestampToAgo($vote->date) => $posX + 120,
            
            );

            if($vote->playerlogin == $player->login) {
                $showOwnIcon = new Quad_Icons128x128_1();
                $voteFrame->add($showOwnIcon);
                $showOwnIcon->setSize(3, 3);
                $showOwnIcon->setY(-0.1);
                $showOwnIcon->setX($posX+2);
                $showOwnIcon->setZ(2);
                $showOwnIcon->setSubStyle($showOwnIcon::SUBSTYLE_ChallengeAuthor);
            }

            $this->maniaControl->manialinkManager
                               ->labelLine($voteFrame, $array);

            $voteFrame->setY($posY);

            $posY -= 4;
            $index++;
        }

        $this->maniaControl->manialinkManager->displayWidget($maniaLink, $player, 'PlayerList');

    }
    public function showRecordsList(array $chat, Player $player) {

        $currentPlayers = $this->maniaControl->playerManager->getPlayers();
        $currentPlayerLogins = array();

        foreach($currentPlayers as $currentPlayer) {
            $currentPlayerLogins[] = $currentPlayer->login;
        }

        $width  = $this->maniaControl->manialinkManager->styleManager->getListWidgetsWidth();
        $height = $this->maniaControl->manialinkManager->styleManager->getListWidgetsHeight();
        $training_mode = $this->maniaControl->settingManager->getSettingValue($this, self::SETTING_TRAINING_MODE);
    
        $records = $this->currentMapRanking;
        
        // create manialink
        $maniaLink = new ManiaLink(ManialinkManager::MAIN_MLID);
        $script    = $maniaLink->getScript();
        $paging    = new Paging();
        $script->addFeature($paging);

        // Main frame
        $frame = $this->maniaControl->manialinkManager->styleManager->getDefaultListFrame($script, $paging);

        $maniaLink->add($frame);

        // Start offsets
        $posX = -$width / 2;
        $posY = $height / 2;

        // Predefine Description Label
        $descriptionLabel = $this->maniaControl->manialinkManager->styleManager->getDefaultDescriptionLabel();

        $frame->add($descriptionLabel);

        // Headline

        $headFrame = new Frame();
        $frame->add($headFrame);
        $headFrame->setY($posY - 5);

        $array = array(
            '#' => $posX + 5, 
            'Time' => $posX + 12,
            'Nickname' => $posX + 25, 
            'Login' => $posX + 50, 
            'Try' => $posX + 70,
            'Recorded on' => $posX + 80, 
            'Date' => $posX + 130, 
        );

        $this->maniaControl->manialinkManager
                           ->labelLine($headFrame, $array);

        $index     = 0;
        $posY      = $height / 2 - 10;
        $pageFrame = null;

        foreach ($records as $record) {

            if($training_mode && $record->playerlogin != $player->login) continue;

            if ($index % 15 === 0) {
                $pageFrame = new Frame();
                $frame->add($pageFrame);
                $posY = $height / 2 - 10;
                $paging->addPage($pageFrame);
            }

            $recordFrame = new Frame();
            $pageFrame->add($recordFrame);

            if ($index % 2 !== 0) {
                $lineQuad = new Quad_BgsPlayerCard();
                $recordFrame->add($lineQuad);
                $lineQuad->setSize($width, 4);
                $lineQuad->setSubStyle($lineQuad::SUBSTYLE_BgPlayerCardBig);
                $lineQuad->setZ(0.001);
            }

            if (strlen($record->nickname) < 2) {
                $record->nickname = $record->login;
            }
            $currentPlayer = false;
            if($record->playerlogin == $player->login) $currentPlayer = true;

            $rank = $record->rank;
            if($training_mode) $rank = '?';


          
            if($training_mode){
                $linecolor = '$39F';
            }elseif($record->rank == 1){
                $linecolor = '$CC0';
            }elseif($record->rank == 2){
                $linecolor = '$999';
            }elseif($record->rank == 3){
                $linecolor = '$963';
            }elseif($record->playerlogin == $player->login){
                $linecolor = '$39F';
            }elseif(in_array($record->playerlogin, $currentPlayerLogins)){
                 $linecolor = '$9CF';
            }elseif($record->serverlogin != $this->serverlogin){
                 $linecolor = '$CCC';

            }else{
                $linecolor = '$FFF';
            }
            $array = array(
                $linecolor . $rank => $posX + 5,
                $linecolor . $this->toSMTime($record->time) => $posX + 12,
                '$fff' . $record->nickname => $posX + 25,
                $record->playerlogin => $posX + 50,
                '$FC0' . $record->try => $posX + 70,
                ($record->name ? $record->name : '$F00Server deleted') => $posX + 80,
                $this->timestampToAgo($record->date) => $posX + 130,
                
            );

            if($record->playerlogin == $player->login) {
                $showOwnIcon = new Quad_Icons128x128_1();
                $recordFrame->add($showOwnIcon);
                $showOwnIcon->setSize(3, 3);
                $showOwnIcon->setY(-0.1);
                $showOwnIcon->setX($posX+2);
                $showOwnIcon->setZ(2);
                $showOwnIcon->setSubStyle($showOwnIcon::SUBSTYLE_ChallengeAuthor);
            }

            $this->maniaControl->manialinkManager
                               ->labelLine($recordFrame, $array);

            $recordFrame->setY($posY);

            $posY -= 4;
            $index++;
        }

        // Render and display xml
        $this->maniaControl->manialinkManager->displayWidget($maniaLink, $player, 'PlayerList');
    }


    /*###########################################################################################################*/
    /*######################################### Core Functions ##################################################*/
    /*###########################################################################################################*/

    public function isAuthConfigured(){
        $password =  $this->maniaControl->settingManager->getSettingValue(get_class(),self::makeServerPasswordSettingName($this->maniaControl),null) ;

        if($password == null OR $password == '') return false;
        else return true;
        
    }
    public function getMapList(){

        return  $this->maniaControl->mapManager->getMaps();

    }

    public function getCurrentMap(){

        return  $this->maniaControl->mapManager->getCurrentMap();

    }

    private function getPlayer($login){

        $player = $this->maniaControl->playerManager->getPlayer($login, true);

        return array('Login' => $player->login, 'NickName' => $player->nickname);

    }

    public function getPlayerList(){


        $players = $this->maniaControl->playerManager->getPlayers();

        if (empty($players)) {
            return array();
        }

        $playerInfo = array();

        foreach ($players as $player) {
            array_push($playerInfo, array('Login' => $player->login, 'NickName' => $player->nickname));
        }

        return $playerInfo;
    }

    public function remoteCall($method,$data = null,$callback = null){

        if($this->callsEnabled == false) return;

        if(!$this->isAuthConfigured()){
            return;
        }

        if(is_callable($data)) {
            $callback = $data;
            $data = null;
        }

       
       
        $serverlogin = $this->serverlogin;

        $password = $this->maniaControl->settingManager->getSettingValue(get_class(),self::makeServerPasswordSettingName($this->maniaControl),null) ;

        $send = array();
        $send['server-login'] = $serverlogin;
        $send['server-password'] = $password;

        $send['data'] = json_encode($data);
        $send = http_build_query($send);


        if(self::CALLS_DEBUG){

            $this->maniaControl->log(self::NAME . 'Call - ' . $method);
        }

        $this->maniaControl->fileReader->postData(self::OBSMANIA_URL . '/remote/' . $method, function ($data, $error) use ($callback) {

                $data = json_decode($data);

                if($error || $data->result == "error"){

                    if(isset($data->code) && $data->code == 'server-auth-fail'){
                        
                        $this->callsEnabled = false;

                        $this->unload();

                    }

                    if($error) $this->maniaControl->chat->sendErrorToAdmins(self::NAME . ' - ' . $error);
                    

                    $this->maniaControl->chat->sendErrorToAdmins(self::NAME . ' Server Error - ' . $data->message);
                    $this->maniaControl->log(self::NAME . ' Server Error - ' . $data->message);

                }
                elseif($data->result == "success"){

                    if($callback != null) $callback($data->data);

                }


        }, $send, false,'application/x-www-form-urlencoded');

    }
    public function writeLocalRecord($playerIndex,$time,$mapIndex){

        $oldRecord = $this->getLocalRecord($mapIndex, $playerIndex);
        if ($oldRecord) {
            if ($oldRecord->time < $time) {
                
                return;
            }
            if ($oldRecord->time == $time) {
                return;
            }
        }
        // Save time
        $mysqli = $this->maniaControl->database->getMysqli();

        $query  = "INSERT INTO `" . self::TABLE_RECORDS . "` (
                `mapIndex`,
                `playerIndex`,
                `time`,
                `checkpoints`
                ) VALUES (
                {$mapIndex},
                {$playerIndex},
                {$time},
                ''
                ) ON DUPLICATE KEY UPDATE
                `time` = VALUES(`time`);";
        $mysqli->query($query);
        if ($mysqli->error) {
            trigger_error($mysqli->error);
            return;
        }
    }

    private function getLocalRecord($mapIndex,  $playerIndex) {

        $mysqli = $this->maniaControl->database
                                     ->getMysqli();

        $query  = "SELECT records.* FROM (
                    SELECT recs.*, @rank := @rank + 1 as `rank` FROM `" . self::TABLE_RECORDS . "` recs, (SELECT @rank := 0) ra
                    WHERE recs.`mapIndex` = {$mapIndex}
                    ORDER BY recs.`time` ASC) records
                WHERE records.`playerIndex` = {$playerIndex};";

        $result = $mysqli->query($query);
        if ($mysqli->error) {
            trigger_error("Couldn't retrieve player record for '{$playerIndex}'." . $mysqli->error);
            return null;
        }
        $record = $result->fetch_object();

        $result->free();

        return $record;

    }

    public function renderText($text,array $vars = array()){

        
        $output = $text;

        foreach($vars as $varname => $var){
             $output = str_replace('{{' . $varname . '}}', $var ,$output);
        }

        return $output;

    }

    public function getVoteColor($vote){

        if($vote <= 5 && $vote >= 4){
            return '$690';
        }elseif($vote < 4 && $vote >= 3){
            return '$CF0';
        }elseif($vote < 3 && $vote >= 2){
            return '$F90';
        }elseif($vote < 2 && $vote >= 1){
            return '$C30';
        }
    }

    public function formatDecimal($value){
      return number_format($value, 2, ',', '');
    }

    public function timestampToAgo($timestamp){
            $datetime = new \DateTime($timestamp);
            $datetime =  $datetime->getTimestamp();
            return \Carbon\Carbon::createFromTimeStamp($datetime)->diffForHumans();
    }
    public function dateTimeToAgo($datetime){
            $datetime = new \Datetime($datetime);
            $datetime =  $datetime->getTimestamp();
            return \Carbon\Carbon::createFromTimeStamp($datetime)->diffForHumans();
    }

    public static function makeServerPasswordSettingName($maniaControl){

        return 'ObsMania password for ' . $maniaControl->server->getInfo()->login;

    }

    public function hideAll(){

        $this->maniaControl->manialinkManager->hideManialink(self::MLID_RECORDS);
        $this->maniaControl->manialinkManager->hideManialink(self::MLID_MAPINFO);
        $this->maniaControl->manialinkManager->hideManialink(self::MLID_SERVERSWITCH);

    }

    /*##########################################################################################################*/
    /*######################################### Import Script ##################################################*/
    /*##########################################################################################################*/


    public function importManiaControlLocalData(array $callback, Player $player){

        if (!$this->maniaControl->authenticationManager
                                ->checkRight($player, AuthenticationManager::AUTH_LEVEL_SUPERADMIN)
        ) {
            $this->maniaControl->authenticationManager
                               ->sendNotAllowed($player);
            return;
        }

        $mysqli = $this->maniaControl->database->getMysqli();

        $query  = "SELECT login,nickname FROM mc_players";
        $result = $mysqli->query($query);
        if ($mysqli->error) {
            trigger_error("Couldn't retrieve players : " . $mysqli->error);
            return null;
        }
        $playersImport = array();
        while ($player = $result->fetch_object()) {
            $playersImport[] = array(
                'playerlogin' => $player->login,
                'nickname' => $player->nickname,
            );
        }
        $result->free();

        $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - Sending ' . count($playersImport) . ' player(s)...');


        $this->remoteCall('mcImportPlayers',$playersImport,function($playerCount){
            var_dump('ObsMania Import - ' . $playerCount . ' new player(s) added');
            $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - ' . $playerCount . ' new player(s) added');

            $mysqli = $this->maniaControl->database->getMysqli();

            $query  = "SELECT `index`,uid,name,authorLogin FROM mc_maps";
            $result = $mysqli->query($query);

            if ($mysqli->error) {
                trigger_error("Couldn't retrieve maps informations : " . $mysqli->error);
                return null;
            }
            
            $mapsImport = array();
            $mapList = array();
            while ($map = $result->fetch_object()) {
                $mapsImport[] = array(
                    'challengeid' => $map->uid,
                    'name' => $map->name,
                    'author' => $map->authorLogin,
                );
                $mapList[] = $map;
            }
            $result->free();

            $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - Sending ' . count($mapsImport) . ' map(s)...');

            $this->remoteCall('mcImportMaps',$mapsImport,function($mapCount) use ($mapList) {

                 $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - ' . $mapCount . ' new map(s) added');


                $this->recursiveRecordImport(0,$mapList);
                

            });
            

        }); 
        
       
    }
    public function toSMTime($time){
        return  substr( Formatter::formatTime($time),0,-1);   
    }

    public function recursiveRecordImport($targetKey, $mapList){

        $mysqli = $this->maniaControl->database->getMysqli();

        if(isset($mapList[$targetKey])){

            $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - Import records of ' . $mapList[$targetKey]->name);


            $query  = "SELECT p.login as login, m.uid as uid, r.time as time FROM mc_localrecords r
            LEFT JOIN mc_players p ON p.index = r.playerIndex
            LEFT JOIN mc_maps m ON m.index = r.mapIndex
            WHERE r.mapIndex = " . $mapList[$targetKey]->index;

            $result = $mysqli->query($query);

            if ($mysqli->error) {
                trigger_error("Couldn't retrieve records on the " . $mapList[$targetKey]->name . " : " . $mysqli->error);
                return null;
            }
            
            $recordsImport = array();

            while ($record = $result->fetch_object()) {
                $recordsImport[] = array(
                    'challengeid' => $record->uid,
                    'playerlogin' => $record->login,
                    'time' => $record->time,
                );
        
            }

            $result->free();

            $this->remoteCall('mcImportRecords',$recordsImport,function($recordCount) use ($targetKey,$mapList){

                 $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - ' . $recordCount . ' record(s) added');
                $this->recursiveRecordImport($targetKey + 1,$mapList);
                

            });



         
        }else $this->maniaControl->chat->sendMessageToAdmins('ObsMania Import - Import Successful ');
        

    }


    
}